#include "../common.h"
#include "list.h"

int debug = 0;                                            // debug output level

#define ANY -1
typedef struct {
   struct list_head link;
   int64_t ip;                                            // denied IP
   int32_t port;                                          // denied TCP port
} deny_t;

struct list_head deny_list;

static void add_list( int64_t ip,  int32_t port ) {       // add one rule restrictions
   deny_t *elem = (deny_t*)malloc( sizeof( deny_t ) );
   if( NULL == elem ) ERR( "allocation error: %m\n" );
   if( debug > 1 ) printf( "add %p rule: %lld:%d\n", elem, ip, port );
   list_add_tail( &elem->link, &deny_list );
   elem->ip = ip;
   elem->port = htons( port ); 
}

static void del_all( void ) {                             // delete all content
   while( !list_empty( &deny_list ) ) {
      deny_t *head = list_first_entry( &deny_list, deny_t, link );
      if( debug > 1 ) printf( "del %p rule: %lld:%d\n", head, head->ip, head->port );
      list_del( deny_list.next );
      free( head );
   }               
}

int find( int64_t ip, int32_t port ) {
   struct list_head *iter;
   list_for_each( iter, &deny_list ) {
      deny_t *elem = list_entry( iter, deny_t, link );
      if( ( elem->ip != ANY && elem->ip == ip ) ||
          ( elem->port != ANY && elem->port == port ) ) {
         if( debug > 1 ) printf( "found in position %p: %lld:%d\n", elem, elem->ip, ntohs( elem->port ) ); 
         return 1;
      }
   };
   if( debug > 1 ) printf( "not found in deny list\n" );
   return 0;
}

char* time_mark( void );
char* policy2str( policy_mess_t pol );                    // show request structure

int main( int argc, char *argv[] ) {
   char dev[ 80 ] = "/dev/", c, sopt[] = "vw:h:p:";
   int df, res;
   useconds_t wait = 0; 
   INIT_LIST_HEAD( &deny_list );
   while( -1 != ( c = getopt( argc, argv, sopt ) ) )
      switch( c ) {
         case 'v':                                        // level o debuging ouyput
            debug++;
            break;
         case 'w':                                        // pause before reply
            if( atoi( optarg ) > 0 ) wait = atoi( optarg ) * 1000;
            break;
         case 'h': {                                       // IP address
               in_addr_t ipn = inet_addr( optarg );
               if( INADDR_NONE == ipn ) {
                  del_all();
                  ERR( "illegal IP address: %s\n", optarg );
               }
               add_list( ipn, ANY );
            };
            break;
         case 'p':                                        // TCP port
            if( atol( optarg ) > 0 ) add_list( ANY, atol( optarg ) );
            break;
         default :
            ERR( "use: %s [-v[v...]] [-h<IP>] [-p<port>] [-w:<milliseconds>]\n", argv[ 0 ] );
      }
   strcat( dev, DEVNAME );
   if( ( df = open( dev, O_RDWR ) ) < 0 ) ERR( "open device %s error: %m\n", dev );
   if( wait != 0 ) {
      res = usleep( wait );
      if( res != 0 ) ERR( "illega wait value %lu: %m\n", (long)wait );
      if( debug ) printf( "delay before answer %lu msec.\n", (long)wait / 1000  );
   }
   while( 1 ) {                                                     // the duty cycle of the policy daemon 
      policy_mess_t policy;
      policy_perm_t permission;
      res = read( df, &policy, sizeof( policy_mess_t ) );
      if( res < 0 ) {
         fprintf( stderr, "read device error: %m\n" );
         break;
      }
      if( sizeof( policy_mess_t ) == res ) {
         if( debug ) printf( "[%s] %s\n", time_mark(), policy2str( policy ) );
         else printf( "%s\n", policy2str( policy ) ); 
      }
      else 
	 printf( "read %d bytes\n", res );
      if( wait != 0 ) usleep( wait );                       // pause in msec. before reply
      permission.sequence = policy.sequence;
      permission.type = find( policy.ipn, policy.portn ) ? DENY : ALLOW;
      res = write( df, &permission, sizeof( policy_perm_t ) );
      if( res != sizeof( policy_perm_t ) ) {
         fprintf( stderr, "write %d bytes, device error: %m\n", res );
         continue;
      }
      if( debug ) 
         printf( "[%s] write %d bytes: %s\n", time_mark(), res, ( ALLOW == permission.type ? "allow" : "deny" ) );
      else
         printf( "write %d bytes: %s\n", res, ( ALLOW == permission.type ? "allow" : "deny" ) );
   };
   del_all();
   close( df );
   return EXIT_SUCCESS;
};

//---------------------------------------------------------------------------------------------------

char* policy2str( policy_mess_t pol ) {
   static char msg[ LEN_MSG ];
   sprintf( msg, "%s: sequence %04d | ", PHASE( pol.type ), pol.sequence );
   struct in_addr ip = { pol.ipn };
   sprintf( msg + strlen( msg ), "IP=%15s | TCP PORT=%5u | %s",
            inet_ntoa( ip ), ntohs( pol.portn ),
            pol.text  );   
   return msg;
}

/*   struct timeb {
               time_t         time;
               unsigned short millitm;
               short          timezone;
               short          dstflag;
           }; */
/*         struct tm {
               int tm_sec;         // seconds 
               int tm_min;         // minutes 
               int tm_hour;        // hours 
               int tm_mday;        // day of the month 
               int tm_mon;         // month 
               int tm_year;        // year 
               int tm_wday;        // day of the week 
               int tm_yday;        // day in the year 
               int tm_isdst;       // daylight saving time 
           }; */
char* time_mark( void ) {
   static char ret[ 10 ];
   struct timeb tb;
   ftime( &tb );
   struct tm *lt = localtime( &tb.time );
   sprintf( ret, "%02d:%02d.%03d", lt->tm_min, lt->tm_sec, tb.millitm % 1000 );
   return ret;
}

